import logging
from pathlib import Path

from makefun import wraps

logger = logging.getLogger(__name__)


def _save_as_joblib(result, output_path):
    # import here to optimize CLIs / Fire
    import joblib

    if output_path.is_dir():
        output_path = output_path / "result.joblib"
    joblib.dump(result, output_path)


class PipelineCLI:
    def __init__(self, output_path=None, transforms=[], store_methods={}):
        self._output_path = output_path
        self._result = None

        for transform in transforms:
            self._add_transform(transform)

        self._store_methods = store_methods

    def _add_transform(self, func):
        @wraps(func)
        def wrapper(*iargs, **kwargs):
            args = []
            for arg in iargs:
                if (arg == ...) or (arg == "..."):
                    args.append([self._result])
                elif arg == "*...":
                    args.append(self._result)
                else:
                    args.append([arg])
            args = sum(args, [])

            for key, value in kwargs.items():
                if value in [..., "..."]:
                    kwargs[key] = self._result

            _func = self._decorate(func)
            self._result = _func(*args, **kwargs)

            if hasattr(self._result, "__name__"):
                if self._result.__name__ == "_merge_overrides_or_call":
                    # if the result comes from an omegaconf-decorated function,
                    # self._result contains the function itself, so it
                    # must be called
                    self._result = self._result()

            return self

        setattr(self, func.__name__, wrapper)

    def __str__(self):
        """Hijack __str__ to store the final result of a pipeline, stored in
        self._result."""

        if self._output_path is None:
            logger.warning(
                "No output path given. Unless one of your steps explicitely "
                "writes to disk, this command will not write any results."
            )
            return ""

        if self._result is None:
            raise ValueError("No result stored on pipeline")

        self._output_path = Path(self._output_path)
        self._output_path.parent.mkdir(parents=True, exist_ok=True)

        if type(self._result) in self._store_methods:
            self._store_methods[type(self._result)](self._result, self._output_path)
        else:
            logger.warning("No custom storing method provided. Storing as .joblib")
            _save_as_joblib(self._result, self._output_path)

        return ""



import re
from itertools import chain
from upath import UPath as Path
from typing import Dict, Optional, Sequence, Union

import numpy as np
import pandas as pd
from torch.utils.data import DataLoader
import pytorch_lightning as pl
from pytorch_lightning.utilities.exceptions import MisconfigurationException
from omegaconf import OmegaConf

from serotiny.io.dataframe import DataframeDataset, read_dataframe
from serotiny.io.dataframe.loaders.abstract_loader import Loader


class ManifestDatamodule(pl.LightningDataModule):
    """A pytorch lightning datamodule based on manifest files. It can either
    use a single manifest file, which contains a column based on which a train-
    val-test split can be made; or it can use three manifest files, one for
    each fold (train, val, test).

    Additionally, if it is only going to be used for prediction/testing, a flag
    `just_inference` can be set to True so the splits are ignored and the whole
    dataset is used.

    The `predict_datamodule` is simply constructed from the whole dataset,
    regardless of the value of `just_inference`.
    """

    def __init__(
        self,
        path: Union[Path, str],
        loaders: Union[Dict, Loader],
        split_column: Optional[Union[Path, str]] = None,
        columns: Optional[Sequence[str]] = None,
        just_inference: bool = False,
        **dataloader_kwargs,
    ):
        """
        Parameters
        ----------
        path: Union[Path, str]
            Path to a manifest file

        loaders: Union[Dict, Loader]
            Loader specifications for each given split.

        split_column: Optional[str] = None
            Name of a column in the dataset which can be used to create train, val, test
            splits.

        columns: Optional[Sequence[str]] = None
            List of columns to load from the dataset, in case it's a parquet file.
            If None, load everything.

        just_inference: bool = False
            Whether this datamodule will be used for just inference
            (testing/prediction).
            If so, the splits are ignored and the whole dataset is used.

        dataloader_kwargs:
            Additional keyword arguments are passed to the
            torch.utils.data.DataLoader class when instantiating it (aside from
            `shuffle` which is only used for the train dataloader).
            Among these args are `num_workers`, `batch_size`, `shuffle`, etc.
            See the PyTorch docs for more info on these args:
            https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader
        """

        super().__init__()
        self.path = path

        # if only one loader is specified, the same loaders are
        # used for all  folds
        loaders = _parse_loaders(loaders)

        path = Path(path)

        # if path is a directory, we assume it is a directory containing multiple
        # manifests - one per split. otherwise, we assume it is the path to a single
        # manifest file, which is expected to have a
        if path.is_dir():
            self.datasets = _make_multiple_manifest_splits(
                path, loaders, columns, just_inference
            )
        else:
            if split_column is None and not just_inference:
                raise MisconfigurationException(
                    "When using a single manifest file, it must have a "
                    "split column, to use for train-val-test splitting."
                )

            self.datasets = _make_single_manifest_splits(
                path, loaders, split_column, columns, just_inference
            )

        self.just_inference = just_inference
        self.dataloader_kwargs = dataloader_kwargs
        self.dataloaders = dict()

    def make_dataloader(self, split):
        kwargs = dict(**self.dataloader_kwargs)
        kwargs["shuffle"] = kwargs.get("shuffle", True) and split == "train"

        return DataLoader(dataset=self.datasets[split], **kwargs)

    def _get_dataloader(self, split):
        if split not in self.dataloaders:
            self.dataloaders[split] = self.make_dataloader(split)

        return self.dataloaders[split]

    def train_dataloader(self):
        if self.just_inference:
            raise TypeError(
                "This datamodule was configured with `just_inference=True`, "
                "so it doesn't have a train_dataloader and can't be "
                "used for training."
            )
        return self._get_dataloader("train")

    def val_dataloader(self):
        if self.just_inference:
            raise TypeError(
                "This datamodule was configured with `just_inference=True`, "
                "so it doesn't have a train_dataloader and can't be "
                "used for training."
            )

        return self._get_dataloader("val")

    def test_dataloader(self):
        split = "predict" if self.just_inference else "test"
        return self._get_dataloader(split)

    def predict_dataloader(self):
        return self._get_dataloader("predict")


def _get_canonical_split_name(split):
    for canon in ["train", "val", "test", "predict"]:
        if split.startswith(canon) or canon.startswith(split):
            return canon
    raise ValueError


def _make_single_manifest_splits(
    manifest_path, loaders, split_column, columns=None, just_inference=False
):
    dataframe = read_dataframe(manifest_path, columns)
    if not just_inference:
        assert dataframe.dtypes[split_column] == np.dtype("O")

    split_names = dataframe[split_column].unique().tolist()
    if not just_inference:
        assert set(split_names).issubset(
            {"train", "training", "valid", "val", "validation", "test", "testing"}
        )

    if split_column != "split":
        dataframe["split"] = dataframe[split_column].apply(_get_canonical_split_name)

    if not just_inference:
        datasets = {}
        for split in ["train", "val", "test"]:
            datasets[split] = DataframeDataset(
                dataframe.loc[dataframe["split"].str.startswith(split)].copy(),
                loaders[split],
            )

    datasets["predict"] = DataframeDataset(dataframe.copy(), loaders["predict"])
    return datasets


def _make_multiple_manifest_splits(
    split_path, loaders, columns=None, just_inference=False
):
    split_path = Path(split_path)
    datasets = {}
    predict_df = []

    for fpath in chain(split_path.glob("*.csv"), split_path.glob("*.parquet")):
        split = re.findall(r"(.*)\.(?:csv|parquet)", fpath.name)[0]
        split = _get_canonical_split_name(split)
        dataframe = read_dataframe(fpath, required_columns=columns)
        dataframe["split"] = split
        dataset = DataframeDataset(dataframe, loaders=loaders[split])
        if not just_inference:
            datasets[split] = dataset
        predict_df.append(dataframe.copy())

    predict_df = pd.concat(predict_df)
    datasets["predict"] = DataframeDataset(predict_df, loaders=loaders["predict"])

    return datasets


def _dict_depth(d):
    return (
        max((_dict_depth(v) if OmegaConf.is_config(v) else 0) for v in d.values()) + 1
    )


def _parse_loaders(loaders):
    depth = _dict_depth(loaders)
    if depth == 1:
        loaders = {split: loaders for split in ["train", "val", "test", "predict"]}
    elif depth != 2:
        raise ValueError(f"Loaders dict should have depth 1 or 2. Got {depth}")

    for k, v in loaders.items():
        loaders[_get_canonical_split_name(k)] = v

    for k in loaders:
        if isinstance(loaders[k], str):
            assert loaders[k] in loaders
            loaders[k] = loaders[loaders[k]]

    for split in ["train", "val", "test"]:
        if split not in loaders:
            raise ValueError(f"'{split}' missing from loaders dict.")

    if "predict" not in loaders:
        loaders["predict"] = loaders["test"]

    return loaders



import numpy as np
import torch


class SwapAxes:
    def __init__(self, first: int, second: int):
        self.first = first
        self.second = second

    def __call__(self, input: np.ndarray):
        print(f"swapping axes {self.first} and {self.second}: {input.shape}")
        output = np.swapaxes(input, self.first, self.second)
        print(f"swapping axes result: {output.shape}")
        return output


class Permute:
    def __init__(self, target_dims):
        self.target_dims = target_dims

    def __call__(self, img):
        return torch.tensor(img).permute(*self.target_dims)



import os
import sys

from hydra.core.config_search_path import ConfigSearchPath
from hydra.plugins.search_path_plugin import SearchPathPlugin

from serotiny.ml_ops.utils import get_serotiny_project


class SerotinySearchPath(SearchPathPlugin):
    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        if "serotiny" in sys.argv[0]:
            if "-sc" in sys.argv and any(["install" in _arg for _arg in sys.argv]):
                pass
            elif not any(
                [("config-dir" in _arg or "-cd" in _arg) for _arg in sys.argv]
            ):
                # look for a .serotiny file in the current directory. if it exists,
                # it means we're in a serotiny project. the .serotiny file will contain
                # the corresponding python package name, such that we can retrieve
                # the config

                project_name = get_serotiny_project()
                if project_name != "serotiny":
                    search_path.append(
                        provider="serotiny-project",
                        path=f"{os.getcwd()}/{project_name}/config",
                    )
                else:
                    print(
                        "Please specify --config-dir in the command line, "
                        "or run serotiny from within a serotiny project."
                    )
                    sys.exit(-1)

        search_path.append(
            provider="serotiny-searchpath-plugin", path="pkg://serotiny.config"
        )



from .transforms import (
    append_class_weights,
    append_labels_to_integers,
    append_one_hot,
    filter_columns,
    filter_rows,
    make_random_df,
    sample_n_each,
    split_dataframe,
)



from .crop import CropCenter
from .normalize import NormalizeAbsolute, NormalizeMean, NormalizeMinMax
from .pad import ExpandColumns, ExpandTo, PadTo
from .project import Project
from .swap import SwapAxes
